/* Copyright (c) 1992, 1998 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
#include "config.h"
#include "jed-feat.h"

#include <windows.h>
#include <stdio.h>
#include <slang.h>

#include "jdmacros.h"

#include <process.h>
#include <stdlib.h>
#include <string.h>

#ifdef _MSC_VER
# include <direct.h>
#elif defined(__BORLANDC__)
# include <dir.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <time.h>

#include <assert.h>
#include <io.h>
#include <errno.h>

#include "display.h"
#include "sysdep.h"
#include "screen.h"
#include "keymap.h"
#include "hooks.h"
#include "ins.h"
#include "ledit.h"
#include "misc.h"
#include "cmds.h"
#include "sig.h"

#if JED_HAS_SUBPROCESSES
# include "jprocess.h"
#endif

extern HANDLE _SLw32_Hstdin;

static int x_insert_cutbuffer(void);
static void x_region_2_cutbuffer(void);
static void x_warp_pointer(void);
static int msw_system(char *, int *, int *);

static SLang_Intrin_Fun_Type Jed_WinCommon_Table[] =
{
   MAKE_INTRINSIC("x_warp_pointer", x_warp_pointer, VOID_TYPE, 0),
   MAKE_INTRINSIC("x_insert_cutbuffer", x_insert_cutbuffer, INT_TYPE, 0),
   /* Prototype: Integer x_insert_cut_buffer ();
    * Inserts cutbuffer into the current buffer and returns the number
    * of characters inserted.
    */
   MAKE_INTRINSIC("x_copy_region_to_cutbuffer", x_region_2_cutbuffer, VOID_TYPE, 0),
   /*Prototype: Void x_copy_region_to_cutbuffer();*/

   SLANG_END_TABLE
};

int init_tty(void)
{
   SLadd_intrinsic_function ("msw_system", (FVOID_STAR) msw_system,
                             SLANG_INT_TYPE, 3, SLANG_STRING_TYPE,
                             SLANG_INT_TYPE, SLANG_INT_TYPE);

   if ((-1 == SLadd_intrin_fun_table (Jed_WinCommon_Table, "MSWINDOWS"))
       || (-1 == SLdefine_for_ifdef ("MOUSE")))
     return 0;

   if (X_Init_Term_Hook != NULL)
     return (*X_Init_Term_Hook) ();

   SLang_init_tty (7, 1, 0);
   Input_Events[0] = _SLw32_Hstdin;
#if JED_HAS_SUBPROCESSES
   InitializeCriticalSection(&Critical_Section);
#endif
   return 1;
}

void reset_tty (void)
{
   if (Batch) return;

   if (X_Init_Term_Hook != NULL)
     {
	if (X_Reset_Term_Hook != NULL) (*X_Reset_Term_Hook) ();
	return;
     }

   SLang_reset_tty();
}

int sys_System(char *command_line)
{
   return shell_command(command_line);
}

void sys_pause (int ms)
{
   Sleep(ms);
}

void init_signals()
{
}



/* Delete the file NAME.  returns 0 on failure, 1 on sucess
 * Under OS/2 and DOS, unlink()[UNIX] and remove()[ANSI] are equivalent.
 */

int sys_delete_file(char *name)
{
   return(1 + remove(name));
}

static WIN32_FIND_DATA findata;
static HANDLE hsearch;
static char Found_Dir[JED_MAX_PATH_LEN];
static char Found_File[JED_MAX_PATH_LEN];
static int Found_FLen;

static void fixup_name(char *file)
{
   int dir;
   char name[JED_MAX_PATH_LEN];

   strcpy (file, Found_Dir);

   strcpy (name, findata.cFileName);
   dir = (findata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);

   strcat(file, name);
   if (dir) strcat(file, "\\");
}

int sys_findfirst(char *thefile)
{
   char *f, the_path[JED_MAX_PATH_LEN], *file, *f1;
   char *pat;

   file = expand_filename(thefile);
   f1 = f = extract_file(file);
   strcpy (Found_Dir, file);

   strcpy (Found_File, file);
   Found_FLen = strlen(Found_File);

   Found_Dir[(int) (f - file)] = 0;

   strcpy(the_path, file);

   while (*f1 && (*f1 != '*')) f1++;
   if (! *f1)
     {
	while (*f && (*f != '.')) f++;
	strcat(the_path, "*");
     }
   pat = the_path;

   if ((INVALID_HANDLE_VALUE != (hsearch = FindFirstFile (pat, &findata))))
     {
 	fixup_name(file);

	/* stupid Windows filesystem is case-insensitive so if
	 * one searches for makefi*, it will return Makefile.in
	 * between the others, so check for wrong values and reject
	 * them */
	if ((strncmp(file, Found_File, Found_FLen) != 0) &&
	    (!sys_findnext(file)))
	  return 0;

 	strcpy(thefile, file);
 	return(1);
     }
   else
     return 0;
}

int sys_findnext(char *file)
{
   if (FindNextFile (hsearch, &findata))
     {
 	fixup_name(file);

	/* stupid Windows filesystem is case-insensitive so if
	 * one searches for makefi*, it will return Makefile.in
	 * between the others, so check for wrong values and reject
	 * them */
	if (strncmp(file, Found_File, Found_FLen) == 0)
	  return 1;
     }

   FindClose (hsearch);

   return 0;
}

unsigned long sys_file_mod_time(char *file)
{
   struct stat st;
   if (-1 == stat (file, &st))
     return 0;
   return (unsigned long) st.st_mtime;
}


/* returns 0 if file does not exist, 1 if it is not a dir, 2 if it is */
int sys_chmod(char *file, int what, int *mode, short *uid, short *gid)
{
#ifdef _MSC_VER
   struct _stat buf;
#else
   struct stat buf;
#endif
   char ourname[256];
   int len;

   if (what)
     {
#ifdef _MSC_VER
	_chmod(file, *mode);
#else
	chmod(file, *mode);
#endif
	return(0);
     }

   /* strip the trailing backslash if necessary */
   strcpy(ourname, file);
   len = strlen(ourname);
   if (len>1 && ourname[len-1] == SLASH_CHAR && ourname[len-2] != ':')
     ourname[strlen(ourname)-1] = '\0';

   if (
#ifdef _MSC_VER
       _stat(ourname, &buf) < 0
#else
       stat(ourname, &buf) < 0
#endif
       )
     return 0;

   *mode = buf.st_mode & 0777;

   if (buf.st_mode & S_IFDIR) return (2);
   return(1);
}

static void x_warp_pointer (void)
{
}

static void x_region_2_cutbuffer (void)
{
   int nbytes;
   char *dat;
   int i, x;
   HGLOBAL hBuf;
   char *buf;

   dat = make_buffer_substring(&nbytes);
   if (dat == NULL) return;

   OpenClipboard(NULL);

   EmptyClipboard();

   for(i = x = 0; i < nbytes; i++)
     if (dat[i] == '\n') x++;

   hBuf = GlobalAlloc(GHND, x + nbytes + 1);
   buf = (char *) GlobalLock(hBuf);

   for(i = x = 0; i < nbytes; i++)
     {
	if (dat[i] == '\n') buf[x++] = '\r';
	buf[x++] = dat[i];
     }

   GlobalUnlock(hBuf);

   SetClipboardData(CF_TEXT, hBuf);
   CloseClipboard();
}

static int x_insert_cutbuffer (void)
{
   int nbytes = 0;
   char *dat;
   int i, x;
   HGLOBAL hBuf;
   char *buf;

   CHECK_READ_ONLY;

   OpenClipboard(NULL);

   hBuf = GetClipboardData(CF_TEXT);
   CloseClipboard();

   if (hBuf)
     {
	buf = (char *)GlobalLock(hBuf);
	for(i = x = 0; buf[i] != 0; i++)
	  if (buf[i] != '\r') x++;

	nbytes = x;
	dat = SLmalloc (x + 1);
	if (dat != NULL)
	  {
	     for(i = x = 0; buf[i] != 0; i++)
	       if (buf[i] != '\r')
		 dat[x++] = buf[i];

	     dat[x] = 0;

	     ins_chars((unsigned char *) dat, nbytes);
	     SLfree (dat);
	  }
	GlobalUnlock(hBuf);
     }

   return nbytes;
}

static int msw_system(char *command_line, int *nCmdShow, int *wait)
{
   UINT retcode;
   HCURSOR hcur, hcur_old;
   STARTUPINFO si;
   PROCESS_INFORMATION pi;

   memset ((char *) &si, 0, sizeof (STARTUPINFO));
   si.wShowWindow = *nCmdShow;
   si.dwFlags = STARTF_USESHOWWINDOW;
   si.cb = sizeof(si);
   retcode = CreateProcess (NULL, command_line, NULL, NULL,
			    0, NORMAL_PRIORITY_CLASS, NULL,
			    NULL, &si, &pi);
   if (!retcode)
     {
	msg_error ("Unable to create process");
	return 1;
     }

   if (!*wait) return 0;

   hcur = LoadCursor(NULL, IDC_WAIT);
   hcur_old = SetCursor(hcur);
   WaitForSingleObject(pi.hProcess, INFINITE);
   SetCursor(hcur_old);
   return 0;
}

#ifndef __CYGWIN32__
/* popen, pclose function for Win32 -- it seems that standard ones work
 * only for console applications */
# define MAX_POPEN 10
typedef struct
{
   FILE *fp;
   HANDLE hprocess;
}
Popen_Type;

static int Popen_Ptr = 0;
static Popen_Type Popen_Buf[MAX_POPEN];

FILE *msw_popen(char *cmd, char *type)
{
   int fd;
   HANDLE fds0[2], fds1[2], hfd;
   char cmd_line[1024], *shell;
   STARTUPINFO si;
   PROCESS_INFORMATION pi;
   SECURITY_ATTRIBUTES sa;
   FILE *fp;
   char *error_msg = "popen failed";

   if (Popen_Ptr == MAX_POPEN)
     {
	msg_error("too much popen's");
	return NULL;
     }

   if (((*type != 'r') && (*type != 'w')) || (*(type + 1) != 0))
     {
	errno = EINVAL;
	return(NULL);
     }

   sa.nLength = sizeof(SECURITY_ATTRIBUTES);
   sa.bInheritHandle = TRUE;
   sa.lpSecurityDescriptor = NULL;

   if (CreatePipe(&fds0[0], &fds0[1], &sa, 0))
     {
	if (!CreatePipe(&fds1[0], &fds1[1], &sa, 0))
	  {
	     CloseHandle(fds0[0]);
	     CloseHandle(fds0[1]);
	     msg_error(error_msg);
	     return NULL;
	  }
     }
   else
     return NULL;

   si.cb = sizeof(STARTUPINFO);
   si.lpReserved = NULL;
   si.lpReserved2 = NULL;
   si.cbReserved2 = 0;
   si.lpDesktop = NULL;
   si.lpTitle = NULL;
   si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
   si.wShowWindow = SW_MINIMIZE;

   si.hStdInput = fds0[0];
   si.hStdOutput = fds1[1];
   si.hStdError = fds1[1];

   shell = getenv("COMSPEC");
   if (shell == NULL) shell = "command";
   sprintf(cmd_line,"%s /c %s", shell, cmd);

   if (!CreateProcess(NULL, cmd_line, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi))
     {
	msg_error(error_msg);
	CloseHandle(fds0[0]);
	CloseHandle(fds0[1]);
	CloseHandle(fds1[0]);
	CloseHandle(fds1[1]);
	return NULL;
     }

   CloseHandle(fds0[0]);
   CloseHandle(fds1[1]);
   if (*type == 'r')
     {
	CloseHandle(fds0[1]);
	hfd = fds1[0];
     }
   else
     {
	CloseHandle(fds1[0]);
	hfd = fds0[1];
     }

   fd = _open_osfhandle((long) hfd, 0);
   if (fd < 0) return NULL;
   fp = _fdopen(fd, type);
   Popen_Buf[Popen_Ptr].hprocess = pi.hProcess;
   Popen_Buf[Popen_Ptr].fp = fp;
   Popen_Ptr++;

   return fp;
}

int msw_pclose(FILE *fp)
{
   DWORD d = -1;
   int i;

   i = Popen_Ptr - 1;
   while (i >= 0)
     {
	if (Popen_Buf[i].fp == fp)
	  {
	     /* send EOF */
	     fprintf(fp, "\x1a");
	     if (fclose(fp) == EOF) return -1;
	     if (WaitForSingleObject(Popen_Buf[i].hprocess, 1000) == WAIT_TIMEOUT)
	       {
		  d = TerminateProcess(Popen_Buf[i].hprocess, -1);
	       }

	     GetExitCodeProcess(Popen_Buf[i].hprocess, &d);
	     i++;
	     while (i < Popen_Ptr)
	       {
		  Popen_Buf[i - 1] = Popen_Buf[i];
		  i++;
	       }

	     Popen_Ptr--;
	     return d;
	  }
	i--;
     }

   return -1;
}
#endif /* !defined(__CYGWIN32__) */

